#ifndef GENERATOR_GAUSSBURST_HH
#define GENERATOR_GAUSSBURST_HH

#include "DataSource.hh"

namespace generator {

  /**  GaussBurst generates a burst of gaussian noise inside a gaussian
    *  amplitude envelope. The generated waveform is:
    *    x(t) = A * exp(-(t-to)^2/(2*sigma^2)) * grn()
    *  The parameter 'Width' specifies the half-width of the waveform to 
    *  be generated and is specified in sigma. At the default width value
    *  of 6.0, the gaussian envelope has decreased to 1.5x10^-8 of its 
    *  central amplitude.
    */
  class GaussBurst : public DataSource {
  public:
      GaussBurst(void);
      GaussBurst(double Amp, double Sigma=0.0, double wid=6.0);
      ~GaussBurst(void);
      GaussBurst* clone(void) const;
      const char* getDataType(void) const;
      Interval getDuration(void) const;
  private:
      void getSeries(const Time& t0, Interval dT, int N, float* data);
  };

  inline const char*
  GaussBurst::getDataType(void) const {
      return "GaussBurst";
  }

} // namespace generator

#endif  // GENERATOR_GAUSSBURST_HH
