#include <iostream>
#include "TimingTypes.hh"
#include "ccrc.hh"

using namespace std;
using namespace xml;


//______________________________________________________________________________
#define byteswap2(a,b) ((char *)&(a))[0] = ((char *)&(b))[1]; ((char *)&(a))[1] = ((char *)&(b))[0];
#define byteswap4(a,b) ((char *)&(a))[0] = ((char *)&(b))[3]; ((char *)&(a))[1] = ((char *)&(b))[2];((char *)&(a))[2] = ((char *)&(b))[1];((char *)&(a))[3] = ((char *)&(b))[0];
#define byteswap8(a,b) ((char *)&(a))[0] = ((char *)&(b))[7]; ((char *)&(a))[1] = ((char *)&(b))[6];((char *)&(a))[2] = ((char *)&(b))[5];((char *)&(a))[3] = ((char *)&(b))[4];((char *)&(a))[4] = ((char *)&(b))[3]; ((char *)&(a))[5] = ((char *)&(b))[2]; ((char *)&(a))[6] = ((char *)&(b))[1]; ((char *)&(a))[7] = ((char *)&(b))[0];

inline void swap (uint32_t* w)
{
   uint32_t tmp;
   byteswap4(tmp, *w);
   *w = tmp;
}

inline bool littleendian ()
{
   uint32_t test = 0; *(uint8_t*) &test = 1;
   return (test == 1);
}

//______________________________________________________________________________
string hexstring (uint32_t x) 
{
   char buf[128];
   sprintf (buf, "0x%08X", x);
   return string (buf);
}

//______________________________________________________________________________
string hexstring16 (uint32_t x) 
{
   char buf[128];
   sprintf (buf, "0x%04X", x);
   return string (buf);
}

//______________________________________________________________________________
int get_index (const char* p)
{
   int index = 0;
   if (!p) return -1;
   while (*p && (*p != '[')) ++p;
   if (!*p) return -1;
   ++p;
   while (*p && isdigit (*p)) {
      index = 10 * index + (*p++ - '0');
   }
   if (*p == ']') return index; 
   else return -1;
}

//______________________________________________________________________________
void setbit (uint32_t& p, int bit, bool set)
{
   p = (p & ~(1 << bit)) | ((set ? 1 : 0) << bit);
}

void setbits (uint32_t& p, int bit, int num, uint32_t set)
{
   uint32_t mask = ~(0xFFFFFFFF << num);
   
   p = (p & ~(mask << bit)) | ((set & mask) << bit);
}

//______________________________________________________________________________
mfo_fanout_t::mfo_fanout_t() 
: mDelayMeasured(0), mDelayControl(0)
{
}

double mfo_fanout_t::measured_delay() const
{
   return (double)mDelayMeasured / 4.294967296E9;
}

double mfo_fanout_t::advance() const 
{
   return (double)(mDelayControl >> 14) / 268.435456E6;
}

double mfo_fanout_t::used_advance() const 
{
   return (double)(((mDelayControl >> 14) + 2) >> 2) / 67.108864E6;
}

bool mfo_fanout_t::up() const
{
   return (mDelayControl & 0x01) != 0;
}

bool mfo_fanout_t::los() const
{
   return (mDelayControl & 0x02) != 0;
}

bool mfo_fanout_t::delayerr() const
{
   return (mDelayControl & 0x04) != 0;
}

bool mfo_fanout_t::missing() const
{
   return (mDelayControl & 0x08) != 0;
}

int mfo_fanout_t::errcount() const
{
   return (mDelayControl >> 4) & 0x0FF;
}

bool mfo_fanout_t::set_measured_delay (double t)
{
   mDelayMeasured = 4.294967296E9 * t + 0.5;
   return true;
}

bool mfo_fanout_t::set_advance (double t)
{
   uint32_t u = 268.435456E6 * t + 0.5;
   setbits (mDelayControl, 14, 18, u);
   return true;
}

bool mfo_fanout_t::set_up (bool b)
{
   setbit (mDelayControl, 0, b);
   return true;
}

bool mfo_fanout_t::set_los (bool b)
{
   setbit (mDelayControl, 1, b);
   return true;
}

bool mfo_fanout_t::set_delayerr (bool b)
{
   setbit (mDelayControl, 2, b);
   return true;
}

bool mfo_fanout_t::set_missing (bool b)
{
   setbit (mDelayControl, 3, b);
   return true;
}

bool mfo_fanout_t::set_errcount (int c)
{
   setbits (mDelayControl, 4, 8, c);
   return true;
}

//______________________________________________________________________________
slave_basic_t::slave_basic_t()
: mBoardId(0), mBoardSerial(0), mSoftwareId (0), mSoftwareRev (0),
  mGPS (0), mAddr(0), mStatus (0), mConf (0)
{
}

bool slave_basic_t::is_fanout() const
{
   return (mBoardId >> 8) == kSlaveFanoutId;
}

int slave_basic_t::revision() const
{
   uint32_t rev = (mBoardId && 0x000000F0) >> 4;
   return rev >= 10 ? rev - 10 : rev + 6;
}

int slave_basic_t::daughter() const
{
   if (is_fanout()) {
      return kSlaveFanout;
   }
   else {
      uint32_t tp = (mSoftwareId >> 8);
      if (tp == kSlaveDuoToneId) {
	 return kSlaveDuoTone;
      }
      else if (tp == kSlaveComparatorId) {
	 return kSlaveComparator;
      }
      else if (tp == kSlaveXOLockingId) {
	 return kSlaveXOLocking;
      }
      else {
	 return kSlaveGeneric;
      }
   }
}

int slave_basic_t::daughter_revision() const
{
   uint32_t rev = (mSoftwareId && 0x000000F0) >> 4;
   return rev >= 10 ? rev - 10 : rev + 6;
}

double slave_basic_t::vcxo() const
{
  return (double)(mStatus >> 16) * 2.5 / 32768.;
}

bool slave_basic_t::sw (int num) const
{
   if ((num >= 1) && (num <= 8)) {
      return mStatus & (1 << (num + 7));
   }
   else if ((num >= 9) && (num <= 10)) {
      return mStatus & (1 << (num - 3));
   }
   else {
      return false;
   }
}

int slave_basic_t::errcount() const
{
   return (mStatus >> 1) & 0x0F;
}

bool slave_basic_t::up() const
{
   return mStatus & 0x01;
}

bool slave_basic_t::los() const
{
   return mStatus & 0x20;
}

int slave_basic_t::crc_errcount() const
{
   return mConf & 0x000000FF;
}


//______________________________________________________________________________
slave_t::slave_t()
{
   memset (mExtended, 0, sizeof (mExtended));
}

uint32_t slave_t::CRC() const
{
   CCRC32::crc_type cval = 0xffffffff;
   slave_t slave;
   memcpy (&slave, (const char*)this, sizeof (slave_t));
   if (littleendian()) {
      uint32_t* l = (uint32_t*)&slave;
      for  (int i = 0; i < (int)sizeof (slave_t) / 4; ++i) swap (l + i);
   }
   CCRC32 crc (0xffffffff, false, false);
   cval = crc.FullCRC ((const char*)&slave, sizeof (slave_t));
   return cval;
}


//______________________________________________________________________________
bool slave_comparator_t::has_ext (int i) const
{
   if ((i >= 0) && (i < 7)) {
      return mExtended[0] & (1 << i);
   }
   else {
      return 0.0;
   }
}


double slave_comparator_t::ext_delay (int i) const
{
   if ((i >= 0) && (i < 7)) {
      return (double)(int32_t)mExtended[i+1] / 4.294967296E9;
   }
   else {
      return 0.0;
   }
}



//______________________________________________________________________________
bool slave_xolocking_t::has_ocxo() const
{
   return mExtended[3] & 0x10000;
}

bool slave_xolocking_t::ocxo_locked() const
{
   return mExtended[3] & 0x20000;
}

double slave_xolocking_t::ocxo() const
{
   return (double)(mExtended[3] & 0x0FFFF) * 5. / 32768.;
}

double slave_xolocking_t::ocxo_delay() const
{
   return (double)(int32_t)mExtended[2] / 4.294967296E9;
}

double slave_xolocking_t::freq_set() const
{
   return (double)mExtended[0];
}

double slave_xolocking_t::freq_meas() const
{
   return (double)mExtended[1];
}


//______________________________________________________________________________
mfo_t::mfo_t ()
: mBoardId(), mBoardSerial(0), mSoftwareId(0), mSoftwareRev(0),
  mGPS(0), mAddr(0), mStatus(0), mConf(0), mOCXOControl(0), mOCXOError(0),
  mDelayUplink(0), mDelayExt(0), mDelayGPS(0), mFanoutUpLos(0), 
  mFanoutMissDerr(0), mFlags (0), mCRC (0)
{
   memset (mGPSStatus, 0, sizeof (mGPSStatus));
   memset (mExtended, 0, sizeof (mExtended));
}

int mfo_t::revision() const
{
   uint32_t rev = (mBoardId && 0x000000F0) >> 4;
   return rev >= 10 ? rev - 10 : rev + 6;
}

double mfo_t::vcxo() const
{
  return (double)(mStatus >> 16) * 2.5 / 32768.;
}

bool mfo_t::sw (int num) const
{
   if ((num >= 1) && (num <= 8)) {
      return mStatus & (1 << (num + 7));
   }
   else if ((num >= 9) && (num <= 10)) {
      return mStatus & (1 << (num - 3));
   }
   else {
      return false;
   }
}

int mfo_t::errcount() const
{
   return (mStatus >> 1) & 0x0F;
}

bool mfo_t::up() const
{
   return mStatus & 0x01;
}

bool mfo_t::los() const
{
   return mStatus & 0x20;
}

bool mfo_t::has_ocxo() const
{
   return mConf & 0x8000;
}

bool mfo_t::has_ext() const
{
   return mConf & 0x4000;
}

bool mfo_t::has_gps() const
{
   return mConf & 0x2000;
}

bool mfo_t::gps_locked() const
{
   return mConf & 0x1000;
}

bool mfo_t::ocxo_locked() const
{
   return mConf & 0x0800;
}

bool mfo_t::use_ext() const
{
   return mConf & 0x0400;
}

bool mfo_t::use_gps() const
{
   return mConf & 0x0200;
}

bool mfo_t::use_uplink() const
{
   return mConf & 0x0100;
}

bool mfo_t::is_master() const
{
   return mConf & 0x0001;
}

bool mfo_t::has_fanout() const
{
   return mConf & 0x0002;
}

int mfo_t::fanout_ports() const
{
   int num = (mConf >> 2) & 0x0F;
   if (has_fanout()) {
      return (num == 0) ? 16 : num;
   }
   else {
      return 0;
   }
}

int mfo_t::crc_err (int port) const
{
   if (port == -1) {
      return mFlags & 0x00FF;
   }
   else if ((port >= 0) && (port < 16)) {
      return (mExtended[port / 4] >> (8*(port % 4))) & 0x00FF;
   }
   else {
      return -1;
   }
}

bool mfo_t::gps_err () const
{
   return (mFlags >> 11) & 0x0001;
}

int mfo_t::gps_errcount () const
{
   return (mFlags >> 8) & 0x0007;
}

double mfo_t::ocxo() const
{
  return (double)(mOCXOControl & 0x0FFFF) * 5. / 32768. - 5.0;
}

double mfo_t::ocxo_delay() const
{
   return (double)mOCXOError / 4.294967296E9;
}

double mfo_t::uplink_delay() const
{
   return (double)mDelayUplink / 4.294967296E9;
}

double mfo_t::gps_delay() const
{
   return (double)mDelayGPS / 4.294967296E9;
}

double mfo_t::ext_delay() const
{
   return (double)mDelayExt / 4.294967296E9;
}

bool mfo_t::set_ocxo(double v)
{
   setbits (mOCXOControl, 0, 16, 3276.8 * (v + 10.0) + 0.5);
   return true;
}

bool mfo_t::set_ocxo_delay(double t)
{
   mOCXOError = 4.294967296E9 * t + 0.5;
   return true;
}

bool mfo_t::set_uplink_delay(double t)
{
   mDelayUplink = 4.294967296E9 * t + 0.5;
   return true;
}

bool mfo_t::set_gps_delay(double t)
{
   mDelayGPS = 4.294967296E9 * t + 0.5;
   return true;
}

bool mfo_t::set_ext_delay(double t)
{
   mDelayExt = 4.294967296E9 * t + 0.5;
   return true;
}

bool mfo_t::fanout_los (int node) const
{
   if ((node < 0) || (node >= 16)) return false;
   else return (mFanoutUpLos & (1 << node));
}

bool mfo_t::fanout_up (int node) const
{
   if ((node < 0) || (node >= 16)) return false;
   else return (mFanoutUpLos & (1 << (node + 16)));
}

bool mfo_t::fanout_missing (int node) const
{
   if ((node < 0) || (node >= 16)) return false;
   else return (mFanoutMissDerr & (1 << (node + 16)));
}

bool mfo_t::fanout_derror (int node) const
{
   if ((node < 0) || (node >= 16)) return false;
   else return (mFanoutMissDerr & (1 << node));
}

bool mfo_t::set_fanout_los (int node, bool val)
{
   if ((node < 0) || (node >= 16)) return false;
   setbit (mFanoutUpLos, node, val);
   return true;
}

bool mfo_t::set_fanout_up (int node, bool val)
{
   if ((node < 0) || (node >= 16)) return false;
   setbit (mFanoutUpLos, node + 16, val);
   return true;
}

bool mfo_t::set_fanout_missing (int node, bool val)
{
   if ((node < 0) || (node >= 16)) return false;
   setbit (mFanoutMissDerr, node + 16, val);
   return true;
}

bool mfo_t::set_fanout_derror (int node, bool val)
{
   if ((node < 0) || (node >= 16)) return false;
   setbit (mFanoutMissDerr, node, val);
   return true;
}

double mfo_t::gps_latitude() const
{
   return (double)((int)mGPSStatus[0]) / 3.6E6;
}

double mfo_t::gps_longitude() const
{
   return (double)((int)mGPSStatus[1]) / 3.6E6;
}

double mfo_t::gps_height() const
{
   return (double)((int)mGPSStatus[2]) / 100.0;
}

double mfo_t::gps_3d_speed() const
{
   return (double)(mGPSStatus[3] >> 16) / 100.0;
}

double mfo_t::gps_2d_speed() const
{
   return (double)(mGPSStatus[3] & 0xFFFF) / 100.0;
}

double mfo_t::gps_heading() const
{
   return (double)(mGPSStatus[4] >> 16) / 10.0;
}

double mfo_t::gps_dop() const
{
   return (double)(mGPSStatus[4] & 0xFFFF) / 100.0;
}

int mfo_t::gps_vis_satellites() const
{
   return (int)(mGPSStatus[5] >> 24);
}

int mfo_t::gps_track_satellites() const
{
   return (int)((mGPSStatus[5] >> 16) & 0xFF);
}

int mfo_t::gps_status() const
{
   return (int)(mGPSStatus[5] & 0xFFFF);
}

std::string mfo_t::gps_fix() const
{
   switch ((gps_status() >> 13) & 0x7) {
      case 7: return "3D Fix";
      case 6: return "2D Fix";
      case 5: return "Propagate Mode";
      case 4: return "Position Hold";
      case 3: return "Acquiring Satellites";
      case 2: return "Bad Geometry";
      default: return "Reserved";
   }
}

bool mfo_t::gps_narrowband() const
{
   return gps_status() & 0x0400;
}

bool mfo_t::gps_antenna() const
{
   return (gps_status() & 0x0006) == 0;
}

std::string mfo_t::gps_serial() const
{
   char buf[8];
   memset (buf, 0, sizeof (buf));
   buf[0] = (mGPSStatus[6] >> 8) & 0xFF;
   buf[1] = (mGPSStatus[6]) & 0xFF;
   buf[2] = (mGPSStatus[7] >> 24) & 0xFF;
   buf[3] = (mGPSStatus[7] >> 16) & 0xFF;
   buf[4] = (mGPSStatus[7] >> 8) & 0xFF;
   buf[5] = (mGPSStatus[7]) & 0xFF;
   return string (buf);
}


uint32_t mfo_t::CRCSlave (int node) const
{
   CCRC32::crc_type cval = 0xffffffff;
   if ((node >= 0) && (node < 16)) {
      cval = mSlave[node].CRC();
   }
   return cval;
}


//______________________________________________________________________________
mfo_tree::mfo_tree()
: mParent (0)
{
   memset (mFanout, 0, sizeof (mFanout));
}

mfo_tree::~mfo_tree()
{
   for (int i = 0; i < 16; ++i) {
      delete mFanout[i];
   }   
}

mfo_tree::mfo_tree(const mfo_tree& mfo)
: mParent (0)
{
   memset (mFanout, 0, sizeof (mFanout));
   *this = mfo;
}

mfo_tree::mfo_tree(const mfo_t& mfo, const string& name)
: mName (name), mParent (0)
{
   memset (mFanout, 0, sizeof (mFanout));
   *(mfo_t*)this = mfo;
}

mfo_tree& mfo_tree::operator= (const mfo_tree& mfo)
{
   *(mfo_t*)this = mfo;
   mName = mfo.mName;
   mParent = mfo.mParent;
   for (int i = 0; i < 16; ++i) {
      if (mFanout[i]) delete mFanout[i];
      mFanout[i] = 0;
      if (mfo.mFanout[i]) {
         mFanout[i] = new (nothrow) mfo_tree (*mfo.mFanout[i]);
	 if (mFanout[i]) {
	    mFanout[i]->set_parent (this);
	 }
      }
   }
   return *this;
}

mfo_tree& mfo_tree::operator= (const mfo_t& mfo)
{
   *(mfo_t*)this = mfo;
   mParent = 0;
   for (int i = 0; i < 16; ++i) {
      if (mFanout[i]) delete mFanout[i];
      mFanout[i] = 0;
   }
   return *this;
}

bool mfo_tree::is_root() const
{
   return mParent == 0;
}

bool mfo_tree::is_fanout (int node) const
{
   if ((node < 0) || (node > 15)) {
      return false;
   }
   return mSlave[node].is_fanout() && fanout_up(node);
}

mfo_tree* mfo_tree::parent()
{
   return mParent;
}

const mfo_tree* mfo_tree::parent() const
{
   return mParent;
}

mfo_tree* mfo_tree::fanout (int node)
{
   if ((node < 0) || (node > 15)) {
      return 0;
   }
   return mFanout[node];
}

const mfo_tree* mfo_tree::fanout (int node) const
{
   if ((node < 0) || (node > 15)) {
      return 0;
   }
   return mFanout[node];
}

void mfo_tree::set_parent (mfo_tree* parent)
{
   mParent = parent;
}

bool mfo_tree::set_fanout (int node, mfo_tree* fanout)
{
   if ((node < 0) || (node > 15)) {
      return false;
   }
   if (mFanout[node]) delete mFanout[node];
   mFanout[node] = fanout;
   if (mFanout[node]) {
      mFanout[node]->set_parent (this);
   }      
   return true;
}

bool mfo_tree::set_fanout (int node, const mfo_tree& fanout)
{
   if ((node < 0) || (node > 15)) {
      return false;
   }
   if (mFanout[node]) delete mFanout[node];
   mFanout[node] = new (nothrow) mfo_tree (fanout);
   if (mFanout[node]) {
      mFanout[node]->set_parent (this);
   }      
   return mFanout[node] != 0;
}

int mfo_tree::fanout_node (const mfo_tree& mfo) const
{
   for (int i = 0; i < 16; ++i) {
      if (fanout_up(i) &&
          (mSlave[i].CRC() == 0) &&
          (mSlave[i].mAddr == mfo.mAddr)) {
         return i;
      }
   }
   return -1;
}

bool mfo_tree::set_fanout (const mfo_tree& mfo)
{
   // check slave list
   int node = fanout_node (mfo);
   if (node >= 0) {
      return set_fanout (node, mfo);
   }
   // traverse slave tree
   for (node = 0; node < 16; ++node) {
      if (mFanout[node]) {
         if (mFanout[node]->set_fanout (mfo)) {
	    return true;
	 }
      }
   }
   return false;
}


//______________________________________________________________________________
std::ostream& xsilMfoPort::write (std::ostream &os) const
{
   // begin
   os << xsilDataBegin (fName ? fName : "Port", 
                        "TimingPort", fLevel) << endl;

   // fanout port status
   os << xsilParameter<bool> (portUp, fData->up(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (portLOS, fData->los(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (portMissing, fData->missing(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (portDelayError, fData->delayerr(), 1, fLevel+1) << endl;
   os << xsilParameter<int> (portErrCount, fData->errcount(), 1, fLevel+1) << endl;
   os << xsilParameter<double> (portDelay, "us", 1E6*fData->measured_delay(), 1, fLevel+1) << endl;
   os << xsilParameter<double> (portAdvance, "us", 1E6*fData->advance(), 1, fLevel+1) << endl;
   os << xsilParameter<double> (portUsedAdvance, "us", 1E6*fData->used_advance(), 1, fLevel+1) << endl;

   // end
   os << xsilDataEnd<float> (fLevel);
   return os;
}


//______________________________________________________________________________
std::ostream& xsilSlaveBasic::write (std::ostream &os) const
{
   // GPS time & address
   os << xsilTime (slaveGPS, fData->mGPS, 0, fLevel) << endl;
   os << xsilParameter<int> (slaveAddr, fData->mAddr, 1, fLevel) << endl;
   int ntuple[8];
   for (int i = 0; i < 8; ++i) ntuple[i] = (fData->mAddr >> (28-4*i)) & 0x0F;
   os << xsilParameter<int> (mfoAddrNtuple, *ntuple, 8, fLevel) << endl;
   // Board & software
   os << xsilParameter<int> (slaveBoard, fData->mBoardId, 1, fLevel) << endl;
   os << xsilParameter<string> (slaveBoardHex, hexstring (fData->mBoardId), 1, fLevel) << endl;
   os << xsilParameter<int> (slaveSerial, fData->mBoardSerial, 1, fLevel) << endl;
   os << xsilParameter<int> (slaveProgram, fData->mSoftwareId, 1, fLevel) << endl;
   os << xsilParameter<string> (slaveProgramHex, hexstring(fData->mSoftwareId), 1, fLevel) << endl;
   os << xsilParameter<int> (slaveRevision, fData->mSoftwareRev, 1, fLevel) << endl;
   // Status
   os << xsilParameter<int> (slaveStatus, fData->mStatus, 1, fLevel) << endl;
   os << xsilParameter<string> (slaveStatusHex, hexstring(fData->mStatus), 1, fLevel) << endl;
   os << xsilParameter<bool> (slaveLOS, fData->los(), 1, fLevel) << endl;
   os << xsilParameter<bool> (slaveUp, fData->up(), 1, fLevel) << endl;
   os << xsilParameter<int> (slaveErrC, fData->errcount(), 1, fLevel) << endl;
   bool dip[10];
   for (int i = 1; i <= 10; ++i) dip[i-1] = fData->sw(i);
   os << xsilParameter<bool> (slaveDIP, dip[0], 10, fLevel) << endl;
   os << xsilParameter<double> (slaveVCXOControl, "V", fData->vcxo(), 1, fLevel) << endl;
   // Configuration
   os << xsilParameter<int> (mfoConfiguration, fData->mConf, 1, fLevel) << endl;
   os << xsilParameter<string> (mfoConfigurationHex, hexstring(fData->mConf), 1, fLevel) << endl;
   os << xsilParameter<int> (slaveCRCErrC, fData->crc_errcount(), 1, fLevel);
   
   return os;
}

//______________________________________________________________________________
std::ostream& xsilSlave::write (std::ostream &os) const
{
   // begin
   os << xsilDataBegin (fName ? fName : "Slave", 
                        "TimingSlave", fLevel) << endl;
   // Slave parameters
   os << xsilParameter<bool> (slaveCRCOK, fData->CRC() == 0, 1, fLevel+1) << endl;
			
   // Basic slave parameters
   os << xsilSlaveBasic (fData, fLevel+1) << endl;

   // Extended
   int ext[24];
   for (int i = 0; i < 24; ++i) ext[i] = fData->mExtended[i];
   os << xsilParameter<int> (slaveExtended, *ext, 24, fLevel+1) << endl;
   
   // slave specific
   switch (fData->daughter()) {
      case kSlaveFanout: {
         os << xsilSlaveFanout (fData, fLevel+1) << endl;
         break;
      }
      case kSlaveDuoTone: {
         os << xsilSlaveDuoTone (fData, fLevel+1) << endl;
         break;
      }
      case kSlaveComparator: {
         os << xsilSlaveComparator (fData, fLevel+1) << endl;
         break;
      }
      case kSlaveXOLocking: {
         os << xsilSlaveXOLocking (fData, fLevel+1) << endl;
         break;
      }
      default: {
         os << xsilParameter<string> (slaveType, "Unknown", 1, fLevel+1) << endl;
         break;
      }
   }

   // end
//   os << xsilDataEnd<int> (24, (const int*)fData->mExtended, fLevel+1);
   os << xsilDataEnd<float> (fLevel);
   return os;
}


//______________________________________________________________________________
std::ostream& xsilSlaveFanout::write (std::ostream &os) const
{
   os << xsilParameter<string> (slaveType, "Fanout", 1, fLevel);
   return os;
}


//______________________________________________________________________________
std::ostream& xsilSlaveDuoTone::write (std::ostream &os) const
{
   os << xsilParameter<string> (slaveType, "DuoTone", 1, fLevel);
   return os;
}


//______________________________________________________________________________
std::ostream& xsilSlaveComparator::write (std::ostream &os) const
{
   os << xsilParameter<string> (slaveType, "Comparator", 1, fLevel) << endl;
   bool present[7];
   for (int i = 0; i < 7; ++i) present[i] = fData->has_ext(i);
   os << xsilParameter<bool> (slaveComparatorExtPPSPresent, *present, 7, fLevel);

   for (int i = 0; i < 7; ++i) {
      char buf[256];
      sprintf (buf, "%s[%i]", slaveComparatorExtPPSDelay, i+1);
      os << endl << xsilParameter<double> (buf, "us", 
          1E6 * fData->ext_delay(i), 1, fLevel);
   }
   return os;
}


//______________________________________________________________________________
std::ostream& xsilSlaveXOLocking::write (std::ostream &os) const
{
   os << xsilParameter<string> (slaveType, "XOLocking", 1, fLevel);
   os << endl << xsilParameter<bool> (slaveXOLockingPresent, "", fData->has_ocxo(), 1, fLevel);
   os << endl << xsilParameter<bool> (slaveXOLockingLocked, "", fData->ocxo_locked(), 1, fLevel);
   os << endl << xsilParameter<double> (slaveXOLockingError, "us", 1E6*fData->ocxo_delay(), 1, fLevel);
   os << endl << xsilParameter<double> (slaveXOLockingControl, "V", fData->ocxo(), 1, fLevel);
   os << endl << xsilParameter<int> (slaveXOLockingFreqSet, "Hz", fData->freq_set(), 1, fLevel);
   os << endl << xsilParameter<int> (slaveXOLockingFreqMeas, "Hz", fData->freq_meas(), 1, fLevel);
   return os;
}


//______________________________________________________________________________
std::ostream& xsilMfoTree::write (std::ostream &os) const
{
   // begin
   os << xsilDataBegin (fName ? fName : "Master",
                        "TimingMasterFanout", fLevel) << endl;
			
   // MFO parameters
   os << xsilParameter<string> (mfoName, fData->mName, 1, fLevel+1) << endl;
   // GPS time & address
   os << xsilTime (mfoGPS, fData->mGPS, 0, fLevel+1) << endl;
   os << xsilParameter<int> (mfoAddr, fData->mAddr, 1, fLevel+1) << endl;
   int ntuple[8];
   for (int i = 0; i < 8; ++i) ntuple[i] = (fData->mAddr >> (28-4*i)) & 0x0F;
   os << xsilParameter<int> (mfoAddrNtuple, *ntuple, 8, fLevel+1) << endl;
   // Board & software
   os << xsilParameter<int> (mfoBoard, fData->mBoardId, 1, fLevel+1) << endl;
   os << xsilParameter<string> (mfoBoardHex, hexstring (fData->mBoardId), 1, fLevel+1) << endl;
   os << xsilParameter<int> (mfoSerial, fData->mBoardSerial, 1, fLevel+1) << endl;
   os << xsilParameter<int> (mfoProgram, fData->mSoftwareId, 1, fLevel+1) << endl;
   os << xsilParameter<string> (mfoProgramHex, hexstring(fData->mSoftwareId), 1, fLevel+1) << endl;
   os << xsilParameter<int> (mfoRevision, fData->mSoftwareRev, 1, fLevel+1) << endl;
   // Configuration 
   os << xsilParameter<int> (mfoConfiguration, fData->mConf, 1, fLevel+1) << endl;
   os << xsilParameter<string> (mfoConfigurationHex, hexstring(fData->mConf), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoMaster, fData->is_master(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoFanout, fData->has_fanout(), 1, fLevel+1) << endl;
   os << xsilParameter<int> (mfoPorts, fData->fanout_ports(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoExtPPSPresent, fData->has_ext(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoOCXOPresent, fData->has_ocxo(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoOCXOLocked, fData->ocxo_locked(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoGPSPresent, fData->has_gps(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoGPSLocked, fData->gps_locked(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoUseExtPPS, fData->use_ext(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoUseGPSPPS, fData->use_gps(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoUseUplinkPPS, fData->use_uplink(), 1, fLevel+1) << endl;
   // Status 
   os << xsilParameter<int> (mfoStatus, fData->mStatus, 1, fLevel+1) << endl;
   os << xsilParameter<string> (mfoStatusHex, hexstring (fData->mStatus), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoUp, fData->up(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoLOS, fData->los(), 1, fLevel+1) << endl;
   os << xsilParameter<int> (mfoErrC, fData->errcount(), 1, fLevel+1) << endl;
   bool dip[10];
   for (int i = 1; i <= 10; ++i) dip[i-1] = fData->sw(i);
   os << xsilParameter<bool> (mfoDIP, dip[0], 10, fLevel+1) << endl;
   os << xsilParameter<double> (mfoVCXOControl, "V", fData->vcxo(), 1, fLevel+1) << endl;
   // OCXO
   os << xsilParameter<double> (mfoOCXOControl, "V", fData->ocxo(), 1, fLevel+1) << endl;
   os << xsilParameter<double> (mfoOCXOError, "us", 1E6*fData->ocxo_delay(), 1, fLevel+1) << endl;
   os << xsilParameter<double> (mfoExtPPSDelay, "us", 1E6*fData->ext_delay(), 1, fLevel+1) << endl;
   os << xsilParameter<double> (mfoUplinkDelay, "us", 1E6*fData->uplink_delay(), 1, fLevel+1) << endl;
   os << xsilParameter<double> (mfoGPSDelay, "us", 1E6*fData->gps_delay(), 1, fLevel+1) << endl;
   // Fanout status
   bool fo[16];
   for (int i = 0; i < 16; ++i) fo[i] = fData->fanout_up(i);
   os << xsilParameter<bool> (mfoFanoutUp, *fo, 16, fLevel+1) << endl;
   for (int i = 0; i < 16; ++i) fo[i] = fData->fanout_los(i);
   os << xsilParameter<bool> (mfoFanoutLOS, *fo, 16, fLevel+1) << endl;
   for (int i = 0; i < 16; ++i) fo[i] = fData->fanout_missing(i);
   os << xsilParameter<bool> (mfoFanoutMissing, *fo, 16, fLevel+1) << endl;
   for (int i = 0; i < 16; ++i) fo[i] = fData->fanout_derror(i);
   os << xsilParameter<bool> (mfoFanoutDelayError, *fo, 16, fLevel+1) << endl;
   // Flags
   os << xsilParameter<int> (mfoFlags, fData->mFlags, 1, fLevel+1) << endl;
   os << xsilParameter<int> (mfoCRCErrC, fData->crc_err(-1), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoGPSErr, fData->gps_err(), 1, fLevel+1) << endl;
   os << xsilParameter<bool> (mfoGPSErrC, fData->gps_errcount(), 1, fLevel+1) << endl;

   // Fanout ports
   for (int i = 0; i < 16; ++i) {
      char buf[256];
      sprintf (buf, "Port[%i]", i);
      os << xsilMfoPort (fData->mFO[i], buf, fLevel + 1) << endl;
   }
   
   // GPS   
   int ext[16];
   for (int i = 0; i < 8; ++i) ext[i] = fData->mGPSStatus[i];
   os << xsilParameter<int> (mfoGPSStatus, *ext, 8, fLevel+1) << endl;
   if (fData->is_master()) {
      os << xsilParameter<double> (mfoGPSLatitude, "deg", fData->gps_latitude(), 1, fLevel+1) << endl;
      os << xsilParameter<double> (mfoGPSLongitude, "deg", fData->gps_longitude(), 1, fLevel+1) << endl;
      os << xsilParameter<double> (mfoGPSHeight, "m", fData->gps_height(), 1, fLevel+1) << endl;
      os << xsilParameter<double> (mfoGPSSpeed3D, "m/s", fData->gps_3d_speed(), 1, fLevel+1) << endl;
      os << xsilParameter<double> (mfoGPSSpeed2D, "m/s", fData->gps_2d_speed(), 1, fLevel+1) << endl;
      os << xsilParameter<double> (mfoGPSHeading, "deg", fData->gps_heading(), 1, fLevel+1) << endl;
      os << xsilParameter<double> (mfoGPSDOP, "", fData->gps_dop(), 1, fLevel+1) << endl;
      os << xsilParameter<int> (mfoGPSVisSatellites, fData->gps_vis_satellites(), 1, fLevel+1) << endl;
      os << xsilParameter<int> (mfoGPSTrackSatellites, fData->gps_track_satellites(), 1, fLevel+1) << endl;
      os << xsilParameter<int> (mfoGPSReceiver, fData->gps_status(), 1, fLevel+1) << endl;
      os << xsilParameter<string> (mfoGPSReceiverHex, hexstring16(fData->gps_status()), 1, fLevel+1) << endl;
      os << xsilParameter<string> (mfoGPSFix, fData->gps_fix(), 1, fLevel+1) << endl;
      os << xsilParameter<bool> (mfoGPSNarrowBand, fData->gps_narrowband(), 1, fLevel+1) << endl;
      os << xsilParameter<bool> (mfoGPSAntenna, fData->gps_antenna(), 1, fLevel+1) << endl;
      os << xsilParameter<string> (mfoGPSSerial, fData->gps_serial(), 1, fLevel+1) << endl;
   }

   // Extended 
   for (int i = 0; i < 8; ++i) ext[i] = fData->mExtended[i];
   os << xsilParameter<int> (mfoExtended, *ext, 8, fLevel+1) << endl;
   for (int i = 0; i <16; ++i) ext[i] = fData->crc_err(i);
   os << xsilParameter<int> (mfoFOCRCErrorC, *ext, 16, fLevel+1) << endl;
   
   // CRC
   os << xsilParameter<int> (mfoCRC, fData->mCRC, 1, fLevel+1) << endl;

   // List of slave modules
   for (int i = 0; i < 16; ++i) {
      char buf[256];
      sprintf (buf, "Slave[%i]", i);
      os << xsilSlave (fData->mSlave[i], buf, fLevel + 1) << endl;
   }
   
   // Fanout modules hanging off this MFO
   for (int i = 0; i < 16; ++i) {
      if (fData->is_fanout (i) && fData->fanout(i)) {
	 char buf[256];
	 sprintf (buf, "Fanout[%i]", i);
         os << xsilMfoTree (fData->fanout(i), buf, fLevel + 1) << endl;
      }
   }
   // end
   os << xsilDataEnd<float> (fLevel);
   return os;
}

//______________________________________________________________________________
std::ostream& xsilMfoTreeList::write (std::ostream &os) const
{
   os << xsilHeader() << endl;
   if (fData) {
      int j = 0;
      for (mfo_tree_list::const_iterator i = fData->begin(); 
           i != fData->end(); ++i) {
	 char buf[256];
	 sprintf (buf, "Master[%i]", j++);
         os << xsilMfoTree (*i, buf) << endl;
      }
   }
   os << xsilTrailer();
   return os;
}

//______________________________________________________________________________
xsilHandlerMfoFanout::xsilHandlerMfoFanout (mfo_fanout_t& object, 
                     const attrlist *attr, bool ignore)
: fObject (&object)
{
}

xsilHandlerMfoFanout::~xsilHandlerMfoFanout()
{
}

bool xsilHandlerMfoFanout::HandleParameter(const std::string& name, 
                        const attrlist& attr,
                        const int& p, int N)
{
   if (strcasecmp (name.c_str(), portErrCount) == 0) {
      fObject->set_errcount (p);
      return true;
   } 
   else 
      return false;
}

bool xsilHandlerMfoFanout::HandleParameter (const std::string& name,
                        const attrlist& attr,
                        const bool& p, int N)
{
   if (strcasecmp (name.c_str(), portLOS) == 0) {
      fObject->set_los (p);
      return true;
   } 
   else if (strcasecmp (name.c_str(), portUp) == 0) {
      fObject->set_up (p);
      return true;
   } 
   else if (strcasecmp (name.c_str(), portDelayError) == 0) {
      fObject->set_delayerr (p);
      return true;
   } 
   else if (strcasecmp (name.c_str(), portMissing) == 0) {
      fObject->set_missing (p);
      return true;
   } 
   else 
      return false;
}
			
bool xsilHandlerMfoFanout::HandleParameter(const std::string& name, 
                        const attrlist& attr, 
                        const double& p, int N)
{
   if (strcasecmp (name.c_str(), portDelay) == 0) {
      fObject->set_measured_delay (1E-6 * p);
      return true;
   } 
   else if (strcasecmp (name.c_str(), portAdvance) == 0) {
      fObject->set_advance (1E-6 * p);
      return true;
   } 
   else
      return false;
}

//______________________________________________________________________________
xsilHandlerSlave::xsilHandlerSlave (slave_t& object, 
                     const attrlist *attr, bool ignore)
: fObject (&object)
{
}

xsilHandlerSlave::~xsilHandlerSlave()
{
}

bool xsilHandlerSlave::HandleParameter (const std::string& name,
                        const attrlist& attr,
                        const bool& p, int N)
{
   return false;
}

bool xsilHandlerSlave::HandleParameter(const std::string& name,
                        const attrlist& attr,
                        const int& p, int N)
{
  if (strcasecmp (name.c_str(), slaveAddr) == 0) {
      fObject->mAddr = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), slaveBoard) == 0) {
      fObject->mBoardId = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), slaveSerial) == 0) {
      fObject->mBoardSerial = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), slaveProgram) == 0) {
      fObject->mSoftwareId = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), slaveRevision) == 0) {
      fObject->mSoftwareRev = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), slaveStatus) == 0) {
      fObject->mStatus = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), slaveConfiguration) == 0) {
      fObject->mConf = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), slaveExtended) == 0) {
      const int* pp = &p;
      for (int i = 0; i < 24 && i < N; ++i) fObject->mExtended[i] = pp[i];
      return true;
   } 
   else
      return false;
}

bool xsilHandlerSlave::HandleParameter(const std::string& name, 
                        const attrlist& attr, 
                        const double& p, int N)
{
   return false;
}

bool xsilHandlerSlave::HandleTime (const std::string& name, 
                        const attrlist& attr,
                        unsigned long sec, unsigned long nsec)
{
  if (strcasecmp (name.c_str(), slaveGPS) == 0) {
      fObject->mGPS = sec;
      return true;
   } 
   else
      return false;
}

//______________________________________________________________________________
xsilHandlerMfoTree::xsilHandlerMfoTree (mfo_tree& object, 
                     const attrlist *attr, bool ignore)
: fObject (&object) 
{
}

xsilHandlerMfoTree::~xsilHandlerMfoTree()
{
}

bool xsilHandlerMfoTree::HandleParameter (const std::string& name,
                        const attrlist& attr,
                        const bool& p, int N)
{
  if (strcasecmp (name.c_str(), mfoFanoutUp) == 0) {
      const bool* pp = &p;
      for (int i = 0; i < 16 && i < N; ++i) 
         fObject->set_fanout_up (i, pp[i]);
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoFanoutLOS) == 0) {
      const bool* pp = &p;
      for (int i = 0; i < 16 && i < N; ++i) 
         fObject->set_fanout_los (i, pp[i]);
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoFanoutMissing) == 0) {
      const bool* pp = &p;
      for (int i = 0; i < 16 && i < N; ++i) 
         fObject->set_fanout_missing (i, pp[i]);
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoFanoutDelayError) == 0) {
      const bool* pp = &p;
      for (int i = 0; i < 16 && i < N; ++i) 
         fObject->set_fanout_derror (i, pp[i]);
      return true;
   } 
   else
      return false;
}

bool xsilHandlerMfoTree::HandleParameter(const std::string& name,
                        const attrlist& attr,
                        const int& p, int N)
{
  if (strcasecmp (name.c_str(), mfoAddr) == 0) {
      fObject->mAddr = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoBoard) == 0) {
      fObject->mBoardId = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoSerial) == 0) {
      fObject->mBoardSerial = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoProgram) == 0) {
      fObject->mSoftwareId = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoRevision) == 0) {
      fObject->mSoftwareRev = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoStatus) == 0) {
      fObject->mStatus = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoConfiguration) == 0) {
      fObject->mConf = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoFlags) == 0) {
      fObject->mFlags = p;
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoGPSStatus) == 0) {
      const int* pp = &p;
      for (int i = 0; i < 8 && i < N; ++i) fObject->mGPSStatus[i] = pp[i];
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoExtended) == 0) {
      const int* pp = &p;
      for (int i = 0; i < 8 && i < N; ++i) fObject->mExtended[i] = pp[i];
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoCRC) == 0) {
      fObject->mCRC = p;
      return true;
   } 
   else
      return false;
}

bool xsilHandlerMfoTree::HandleParameter(const std::string& name, 
                        const attrlist& attr, 
                        const double& p, int N)
{
   if (strcasecmp (name.c_str(), mfoOCXOControl) == 0) {
      fObject->set_ocxo (p);
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoOCXOError) == 0) {
      fObject->set_ocxo_delay (1E-6 * p);
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoExtPPSDelay) == 0) {
      fObject->set_ext_delay (1E-6 * p);
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoUplinkDelay) == 0) {
      fObject->set_uplink_delay (1E-6 * p);
      return true;
   } 
   else if (strcasecmp (name.c_str(), mfoGPSDelay) == 0) {
      fObject->set_gps_delay (1E-6 * p);
      return true;
   } 
   else
      return false;
}

bool xsilHandlerMfoTree::HandleParameter(const std::string& name, 
                        const attrlist& attr,
                        const std::string& p)
{
   if (strcasecmp (name.c_str(), mfoName) == 0) {
      fObject->mName = p;
      return true;
   }
   else
      return false;
}

bool xsilHandlerMfoTree::HandleTime (const std::string& name, 
                        const attrlist& attr,
                        unsigned long sec, unsigned long nsec)
{
  if (strcasecmp (name.c_str(), mfoGPS) == 0) {
      fObject->mGPS = sec;
      return true;
   } 
   else
      return false;
}

xsilHandler* xsilHandlerMfoTree::GetHandler (const attrlist& attr)
{
   attrlist::const_iterator ni = attr.find (xmlName);
   attrlist::const_iterator ti = attr.find (xmlType);
   int index = -1;
   if (ni != attr.end()) {
      index = get_index (ni->second.c_str());
      if (index > 15) index = -1;
   }
   if ((index >= 0) && 
      (strcasecmp (ti->second.c_str(), "TimingMasterFanout") == 0)) {
      fObject->set_fanout (index, mfo_tree());
      mfo_tree* fo = fObject->fanout(index);
      if (fo) {
         return new (nothrow) xsilHandlerMfoTree(*fo, &attr);
      }
      else {
         return 0;
      }
   } 
   else if ((index >= 0) &&
      (strcasecmp (ti->second.c_str(), "TimingPort") == 0)) {
      return new (nothrow) xsilHandlerMfoFanout(fObject->mFO[index], &attr);
   }
   else if ((index >= 0) && 
      (strcasecmp (ti->second.c_str(), "TimingSlave") == 0)) {
      return new (nothrow) xsilHandlerSlave(fObject->mSlave[index], &attr);
   }
   else {
      return 0;
   }
}

//______________________________________________________________________________
xsilHandler* xsilHandlerQueryMfoTree::GetHandler (const attrlist& attr)
{
   attrlist::const_iterator ti = attr.find (xmlType);
   if ((ti != attr.end()) && 
      (strcasecmp (ti->second.c_str(), "TimingMasterFanout") == 0)) {
      fObjects->push_back (mfo_tree());
      return new (nothrow) xsilHandlerMfoTree(fObjects->back(), &attr);
   } 
   else {
      return 0;
   }
}

