////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ccrc.hh"

////////////////////////////////////////////////////////////////////////////////

CCRC32::CCRC32(crc_type initalValue, bool reverseBitOrder, bool finalXor)
: mInitialValue (initalValue), mReverseBitOrder (reverseBitOrder), 
  mFinalXor(finalXor)
{
   memset(ulTable, 0, sizeof(ulTable));

   // 256 values representing ASCII character codes.
   for(int iCodes = 0; iCodes <= 0xFF; iCodes++)
   {
      if (mReverseBitOrder) ulTable[iCodes] = Reflect(iCodes, 8) << 24;
                       else ulTable[iCodes] = iCodes << 24;

      for(int iPos = 0; iPos < 8; iPos++)
      {
         ulTable[iCodes] = (ulTable[iCodes] << 1) ^
            (ulTable[iCodes] & (1 << 31) ? CRC32_POLYNOMIAL : 0);
      }

      if (mReverseBitOrder) ulTable[iCodes] = Reflect(ulTable[iCodes], 32);
                       else ulTable[iCodes] = ulTable[iCodes];
   }
}

////////////////////////////////////////////////////////////////////////////////
// Reflection is a requirement for the official CRC-32 standard.
//   You can create CRCs without it, but they won't conform to the standard.

CCRC32::crc_type CCRC32::Reflect(crc_type ulReflect, char cChar)
{
   crc_type ulValue = 0;

   // Swap bit 0 for bit 7 bit 1 For bit 6, etc....
   for(int iPos = 1; iPos < (cChar + 1); iPos++) {
      if(ulReflect & 1) ulValue |= 1 << (cChar - iPos);
      ulReflect >>= 1;
   }
   return ulValue;
}

////////////////////////////////////////////////////////////////////////////////

CCRC32::crc_type CCRC32::FullCRC(const char *sData, crc_size ulLength)
{
   crc_type ulCRC = mInitialValue;
   PartialCRC(&ulCRC, sData, ulLength);
   if (mFinalXor) {
      return ulCRC ^ 0xffffffff;
   }
   else {
      return ulCRC;
   }
}

////////////////////////////////////////////////////////////////////////////////

//For Example usage example, see FileCRC().
void CCRC32::PartialCRC(crc_type *ulInCRC, const char *sData, 
		        crc_size ulLength)
{
   const unsigned char* p = (const unsigned char*) sData;
   if (mReverseBitOrder) {
      *ulInCRC = Reflect (*ulInCRC, 32);
      while(ulLength--) {
         *ulInCRC = (*ulInCRC >> 8) ^ ulTable[(*ulInCRC & 0xFF) ^ *p++];
      }
      *ulInCRC = Reflect (*ulInCRC, 32);
   }
   else {
      while(ulLength--) {
         *ulInCRC = (*ulInCRC << 8) ^ ulTable[((*ulInCRC >> 24) & 0xFF) ^ *p++];
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
