#include <tid.h>
#include <event.h>
#include <string>
#include <TidString.h>
/*

bool TFCoords::operator < (const TFCoords& s) {

  if(s.t > t) return 1;
  if(s.t == t && s.f > f) return 1;
  return 0;

}

bool TFCoords::operator && (const TFCoords& s) { // test if nearest neighbours

  unsigned int d = abs(s.t - t) + abs(s.f - f);

  if(d == 1) return 1;

  if(d == 0) throw TidError ( "Testing two overlapping pixels for being nearest neighbours.");
  

  return 0;
}

*/

//--------------------EVENT----------------------------------

Event::Event () // constructor
{
  type = Unknown;

  maturity = Foetus;

  time = -1;

  duration = -1.0;

  size = 0;
  cap = EventCap;

  t = (unsigned int*)malloc(EventCap * sizeof(unsigned int));
  f = (unsigned int*)malloc(EventCap * sizeof(unsigned int));
  P = (float*)malloc(EventCap * sizeof(float));

  minfreq = maxfreq = 0;
}


void Event::Reset () { // reset values to those of default constructor

  /*
  type = Unknown;

  maturity = Foetus;

  time = -1;

  duration = -1.0;
  */

  size = 0;

  minfreq = maxfreq = 0;
}


/*
Event::Event (unsigned int cid_,
	      const EventType type_,
	      const Time time_, 
	      const Interval duration_,
	      unsigned int t,
	      unsigned int f)
{
  cid = cid_;

  type = type_;

  if(type_ != Diagnostic && type_ != Reference)
    throw TidError ( "Invalid type in Event constructor.");

  maturity = Child;

  time = time_;

  duration = duration_;

  TFCoords tfc;
  tfc.t = t; 
  tfc.f = f;
  tfc.P = 0.0;

  filters.push_back(tfc);

}
*/

Event::Event (unsigned int cid_,
	      const EventType type_,
	      const Time time_, 
	      const Interval duration_,
	      unsigned int t_,
	      unsigned int f_,
	      float P_)
{
  cid = cid_;

  type = type_;

  if(type_ != Diagnostic && type_ != Reference)
    throw TidError ( "Invalid type in Event constructor.");

  maturity = Child;

  time = time_;

  duration = duration_;

  size = 1;

  t[0] = t_;
  f[0] = f_;
  P[0] = P_;

  minfreq = maxfreq = f_;
}


void Event::Set (unsigned int cid_,
		 const EventType type_,
		 const Time time_, 
		 const Interval duration_,
		 unsigned int t_,
		 unsigned int f_,
		 float P_)
{
  cid = cid_;

  type = type_;

  if(type_ != Diagnostic && type_ != Reference)
    throw TidError ( "Invalid type in Event constructor.");

  maturity = Child;

  time = time_;

  duration = duration_;

  size = 1;

  t[0] = t_;
  f[0] = f_;
  P[0] = P_;

  minfreq = maxfreq = f_;
}


/*
Event::Event (unsigned int cid_,
	      const EventType type_,
	      const Time time_) // constructor

{
  cid = cid_;

  type = type_;

  if(type_ != Diagnostic && type_ != Reference)
    throw TidError ( "Invalid type in Event constructor.");

  maturity = Child;

  time = time_;
}
*/

/*
Event::Event (const Event &s)  // copy constructor
{
  type = s.type;
  maturity = s.maturity;
  time = s.time;
  duration = s.duration;
  filters = s.filters;
  status = s.status;
}
*/

Event::~Event () // destructor
{
  free(t);
  free(f);
  free(P);
}


bool Event::operator < (const Event& s) { // to sort elements by decreasing order of LastTimeID

  if(s.t[s.size-1] < t[size-1]) return 1;

return 0;

}


void Event::IncreaseCap () {   // increase cap to 2*size

  cap = 2 * size;

  t = (unsigned int *)realloc(t, cap * sizeof(unsigned int));

  f = (unsigned int *)realloc(f, cap * sizeof(unsigned int));

  P = (float *)realloc(P, cap * sizeof(float));

}


void Event::operator += (Event const &b) // merge two events, preserving the ordering of the elements
{
  if(type != b.type) 
    throw TidError ( "Attempt to merge two events of different types");

  if(maturity != b.maturity) 
    throw TidError ( "Attempt to merge two events of different maturities");

  if(this == &b)
    throw TidError ( "Attempt to merge an event with itself.");

  // merge
  unsigned int j = size;
  size += b.size;
  if(size > cap) IncreaseCap();

  for(unsigned int i=0; i<b.size; i++, j++) {
    t[j] = b.t[i];
    f[j] = b.f[i];
    P[j] = b.P[i];

    if(b.f[i] > maxfreq) maxfreq = b.f[i];
    if(b.f[i] < minfreq) minfreq = b.f[i];
  }

  // check ordering; N^2 sorting algorithm, but fastest for N<20
  int i;
  unsigned int at, af;
  float aP;

  for(j=1;j<size;j++) {

    at = t[j];
    af = f[j];
    aP = P[j];

    i=j-1;
    while(at < t[i] ||
	  (at == t[i] && af < f[i])) {
      t[i+1] = t[i];
      f[i+1] = f[i];
      P[i+1] = P[i];
      i--;
      if(i<0) break;
    }
    t[i+1] = at;
    f[i+1] = af;
    P[i+1] = aP;
  }

  //re-evaluate duration and start time
  if(b.time < time) time = b.time;

  Interval tdur((double)(1 + t[size-1] - t[0]) * tidg->GetTRes());
  if(tdur > duration) duration = tdur;
}


bool Event::operator && (const Event& b){ // test if events touch
  // this routines uses most of the CPU time

  if(type != b.type) 
    throw TidError ( "Attempt to test two events of different types");

  if(maturity != b.maturity) 
    throw TidError ( "Attempt to test two events of different maturities");

  if(this == &b)
    throw TidError ( "Attempt to test an event with itself.");

  
  /*
  if(t[size-1] + 1 < b.t[0] ||
     t[0] > b.t[b.size-1] + 1)
    return 0;
  */

  unsigned int dt, df, i, j;
  for(i=0; i<size; i++)
    for(j=0; j<b.size; j++) {
      dt = abs((int)(t[i]) - (int)(b.t[j]));
      df = abs((int)(f[i]) - (int)(b.f[j]));

      //      if(dt > 1) break; // because time-ordered

      if(df == 0 && dt == 1) return 1;
      if(df == 1 && dt == 0) return 1;
    }
  

  return 0;
}

#define LARGEINT 65535

unsigned int Event::operator , (const Event& s) { // get distance

  unsigned int i,j,d=LARGEINT, tt;

  for(i=0;i<size;i++)
    for(j=0;j<s.size;j++)
      if((tt = abs((int)(t[i]) - (int)(s.t[j])) + abs((int)(f[i]) - (int)(s.f[j]))) < d) d=tt;

  return d;
}


float Event::Power () { // compute total power

  float Po = 0.0;

  for(unsigned int i=0; i<size; i++) Po += P[i];

  return Po;

}


float Event::ExcessPower (vector<float>& PThresh) { // compute total Excess power

  float Po = 0.0;

  for(unsigned int i=0; i<size; i++) Po += P[i] - PThresh[f[i]];

  return Po;

}

TidString* Event::Marshal() { // convert info to string

  TidString* s = new TidString();

  unsigned int cid = GetChannelID();
  SWrite(s,&cid,sizeof(unsigned int));

  switch(type) {
  case Unknown:
    s->append("U");
    break;
  case Diagnostic:
    s->append("D");
    break;
  case Reference:
    s->append("R");
    break;
  default:
    s->append("X");
    TidWarning ("Undefined event type.");
  }

  
  Time::ulong_t ta = time.getS();
  SWrite(s,&ta,sizeof(Time::ulong_t)); 

  ta = time.getN();
  SWrite(s,&ta,sizeof(Time::ulong_t)); 

  long tb =  duration.GetS();
  SWrite(s, &tb, sizeof(long)); 


  ta = duration.GetN();
  SWrite(s, &ta, sizeof(Time::ulong_t)); 


  /*
  unsigned int si = status.size();
  SWrite(s, &si, sizeof(unsigned int)); 

  for(unsigned int i=0; i<status.size(); i++) 
    switch (status[i]) {
    case Undefined:
      s->append("U");
      break;
    case Independent:
      s->append("I");
      break;
    case Coincident:
      s->append("C");
      break;
    default:
      s->append("X");
      TidWarning ("Undefined event status.");
    }
  */

  unsigned int si = 0;
  SWrite(s, &si, sizeof(unsigned int)); 

  si = size;
  SWrite(s, &si, sizeof(unsigned int)); 

  for(unsigned int i=0; i<si; i++) {
    SWrite(s,t+i,sizeof(unsigned int)); 
    SWrite(s,f+i,sizeof(unsigned int)); 
    SWrite(s,P+i,sizeof(float)); 
  }


  return s;
}


void swrite(char* s, void *ptr, size_t siz) { // convert data to byte array
  char* t = (char *)ptr;
  size_t i;
  for(i=0; i<siz; i++) 
    s[i]=t[i];
}

void SWrite(TidString* s, void *ptr, size_t siz) {
  static char buf[256];

  swrite(buf,ptr,siz);

  s->append(buf,siz);

}

void Event::AppendEventToFile (char* file) {

  FILE *out;

  if((out=fopen(file,"a"))==NULL)
    throw TidError ( string("Unable to open file ") + file + string(" for appending.") );

  fwrite(&size, sizeof(unsigned int), 1, out);

  fwrite(t, sizeof(unsigned int), size, out);

  fwrite(f, sizeof(unsigned int), size, out);

  fclose(out);

}

/*
class SortEvent { 
public:
  bool operator() (Event*, Event*);
};

bool SortEvent::operator()(Event* s1, Event* s2) {

  if(s1->LastTimeID() < s2->LastTimeID()) return 1;
  return 0;

}
*/

//--------------------EventsList-------------------------------

EventsList::EventsList () // default constructor
{
  Raw = 0;

  alpha = 1.0;
  sigma = 1;
  delta = NULL;
  maxDt = 2;
  firstTime = 0;
}


/*
EventsList::EventsList (string chnname_,
                        EventType type_,
                        double fLow_,
                        double fHigh_,
                        Interval deltaT_)
{
  Raw = 0;

  sigma = 1;
  delta = NULL;
  alpha = 1.0;
  PThresh = 1.0;

  maxDt = 2;
  minSizeDelta = 1;

  chnname = chnname_;
  type = type_;

  fLow = fLow_;
  fHigh = fHigh_;
  fRes = 1.0 / deltaT_;
  
  deltaT = deltaT_;

  firstTime = 0;
}
*/


EventsList::EventsList (string chnname_,
			unsigned int sigma_,
			vector<unsigned int>& delta_,
			float alpha_,
			vector<float>& pthresh_,
			EventType type_,
			double fLow_,
			double fHigh_,
			Interval deltaT_)
{
  Raw = 0;

  if(delta_.size() != sigma_ * (sigma_ - 1) / 2)
    throw TidError ("Value of sigma is incompatible with length of delta vector in EventsList constructor");

  alpha = alpha_;
  PThresh = pthresh_;

  SetDelta(sigma_, delta_);

  chnname = chnname_;
  type = type_;

  fLow = fLow_;
  fHigh = fHigh_;

  if(fHigh_ < 0.0) useNyquist = 1;
  else useNyquist = 0;

  fRes = 1.0 / deltaT_;
  
  deltaT = deltaT_;

  firstTime = 0;

  nowID = 0;
}

EventsList::~EventsList () // destructor
{
  PThresh.clear();

  list<Event*>::iterator p;

  for(p=eventsList.begin(); p!=eventsList.end(); p++)
    tidg->DeleteEvent(*p); // put back to ePool

  eventsList.clear();

  for(unsigned int i=0;i<sigma-1;i++) free(delta[i]);
  if(delta!=NULL) free(delta);
}


void EventsList::Clear () {

  list<Event*>::iterator p;

  for(p=eventsList.begin(); p!=eventsList.end(); p++)
    tidg->DeleteEvent(*p); // put back to ePool

  eventsList.clear();

}


void EventsList::Add(Event* s){ // add new event at end of list

  eventsList.push_back(s);
  //  nowID = s->LastTimeID();

}


void EventsList::Add(unsigned int cid,
		     const FSpectrum* s) { // add all events from a given Spectrum object

  //ERROR check
  if(s->getLowFreq() != fLow)
    throw TidError ( string("Minimum frequencies of spectrum and events list don't match for channel ") + chnname);

  if(s->getHighFreq() != fHigh) 
    throw TidError ( string("Maximum frequencies of spectrum and events list don't match for channel ") + chnname);

  if(s->getFStep() != fRes)
    throw TidError ( string("Frequency resolutions of spectrum and events list don't match for channel ") + chnname);

  if(s->getEndTime() - s->getStartTime() != deltaT)
    throw TidError ( string("Data stretch duration of spectrum and events list don't match for channel ") + chnname);

  if(s->getStartTime() < firstTime)
    throw TidError ( string("Trying to process data older than latest ones for channel ") + chnname);


  //WARNING check
  unsigned int ntime = (unsigned int)ceil((s->getStartTime() - firstTime) / deltaT);


  const float* data = s->refData(); // pointer to data

  for(unsigned int i=1; i<s->getNStep(); i++) // skip DC 
    if(data[i] >= PThresh[i]) { // we have a winner

      Event* p = tidg->NewEvent();
      p->Set(cid,
	     type,
	     s->getStartTime(),
	     deltaT,
	     ntime,
	     i,
	     data[i]);
      Add(p);

    }
}


list<Event*>::iterator EventsList::Erase(list<Event*>::iterator p) { // remove on element and free Event obj
  
  tidg->DeleteEvent(*p); // put back into ePool

  return eventsList.erase(p); // remove from list

}


void EventsList::SetDelta (const unsigned int sigma_, 
			   const unsigned int* deltaVect) { // set parameters 

  sigma = sigma_;

  delta = (unsigned int**)malloc((sigma-1)*sizeof(unsigned int *));

  minSizeDelta = 1;

  int k = 0;
  for(unsigned int i=0; i<sigma-1; i++) {
    delta[i] = (unsigned int *)malloc((sigma-1-i)*sizeof(unsigned int));
    for(unsigned int j=0;j<sigma-1-i;j++, k++) {
      delta[i][j] = deltaVect[k];
      if(deltaVect[k] > 1) minSizeDelta = i+1;
    }
  }

  maxDt = 2;
  for(unsigned int i=0;i<sigma-1;i++)
    for(unsigned int j=0;j<sigma-1-i;j++)
      if(delta[i][j] > maxDt) maxDt = delta[i][j];

  
  TidMessage("Parameters sigma and delta's were modified.");
}


void EventsList::SetDelta (const unsigned int sigma_, 
			   vector<unsigned int>& deltaVect) { // set parameters 

  sigma = sigma_;

  delta = (unsigned int**)malloc((sigma-1)*sizeof(unsigned int *));

  minSizeDelta = 1;

  int k = 0;
  for(unsigned int i=0; i<sigma-1; i++) {
    delta[i] = (unsigned int *)malloc((sigma-1-i)*sizeof(unsigned int));
    for(unsigned int j=0;j<sigma-1-i;j++, k++) {
      delta[i][j] = deltaVect[k];
      if(deltaVect[k] > 1) minSizeDelta = i+1;
    }
  }

  maxDt = 2;
  for(unsigned int i=0;i<sigma-1;i++)
    for(unsigned int j=0;j<sigma-1-i;j++)
      if(delta[i][j] > maxDt) maxDt = delta[i][j];

  
  TidMessage("Parameters sigma and delta's were modified.");
}



void EventsList::SetAlpha (const float alpha_) {
  
  alpha = alpha_;

  TidMessage("Parameter alpha was modified.");

}


void EventsList::SetPThresh (const vector<float>& P_) {
  
  PThresh = P_;

  TidMessage("Power threshold parameter was modified.");

}


void EventsList::Dump () { // dump to screen

  list<Event*>::iterator p;

  cout<<"Age\tSize\tMaturity\n";

  for(p=eventsList.begin(); p != eventsList.end(); p++) 
    cout<<nowID-(*p)->LastTimeID()<<"\t"<<(*p)->GetSize()<<"\t"<<(*p)->GetMaturity()<<endl;

}


void EventsList::Clean () {
  
  if(!Raw) {
    if(eventsList.size() > 1) {
      Cluster(); 
      CThresh(); 
      CPower();
    }
  }
  else 
    MatureAll();

}


void EventsList::MatureAll () { 

  list<Event*>::iterator p;

  for(p = eventsList.begin(); p != eventsList.end(); p++)
    (*p)->BeMature();

}


void EventsList::Cluster () { // clean the list by merging and deleting, i.e, does nearest neighb clustering analysis

  //NOTE: this is not optimal, but should be good for small lists


  /*
  //DEBUG ONLY!!!!!!!!
  list<Event*>::iterator qqq, ppp;
  for(ppp=qqq=eventsList.begin(), qqq++; qqq!=eventsList.end(); ppp++,qqq++)
    if((*ppp)->LastTimeID() > (*qqq)->LastTimeID()) 
      throw TidError ("List not ordered.");
  */


  list<Event*>::iterator tmp1 = eventsList.end(), tmp2;
  tmp1--; // end is one-past-last element

  tmp2 = tmp1;
  tmp1--;

  //    cout<<"Before merge in freq:\n";
  //    Dump();

  //Merge in frequency:
  for(;;) { // loop over age 0, single pixel events


    //    if((*tmp1)->GetChannelID() == 0)
    //      (*tmp1)->AppendEventToFile("RawData.clust");

    if(nowID != (*tmp1)->LastTimeID()) break;

    if(**tmp1 && **tmp2) // check if they touch
      {
	**tmp1 += **tmp2; // add second to first
	Erase(tmp2); // remove second
    
	if(tmp1==eventsList.begin()) break;
	else {tmp2 = tmp1; tmp1--;} // go previous
      }
    else
      {
	if(tmp1==eventsList.begin()) break;
	tmp1--;
	tmp2--;
      }
  }
  // at this point, tmp1 is first age >=1 event, tmp2 is last age 0 or tmp1 is first element in list, of age 0.

  //   cout<<"After merge in freq:\n";
  //    Dump();
  
  //Merge in time:

  while(nowID - (*tmp1)->LastTimeID() == 1) { // loop over age 1 events in list
    
    for(tmp2 = eventsList.end(), tmp2--; tmp2 != tmp1; tmp2--) { // loop to find overlap in freq
      //      if((*tmp2)->GetMaxFreq() + 1 >= (*tmp1)->GetMinFreq() &&
      //	 (*tmp2)->GetMinFreq() <= (*tmp1)->GetMaxFreq() + 1)
	{

	  if(**tmp1 && **tmp2) // check if they touch
	    {
	      **tmp1 += **tmp2; // fuse them

	      tmp2 = Erase(tmp2); // remove from list, free Event *tmp2
	    }
	}

    }
    
    if(tmp1 == eventsList.begin()) break;

    tmp1--; // go to previous
  }
  // at this point, tmp1 is first age >= 2 event, or tmp2 or tmp1 is first element

  //    cout<<"after merge in time:\n";
  //    Dump();


  // sort the whole list by age

  
  // N^2 sorting algorithm, but fastest for N<20
  list<Event*>::iterator i,j,k; 
  Event* a;

  for(j=eventsList.begin(), j++; j != eventsList.end(); j++) {
    a = *j;
    i=j; i--;
    
    bool fligou=0;

    while(**i < *a) {
      k=i; k++;
      *k = *i;

      if(i == eventsList.begin()) {fligou=1; break;}
      else i--;
    }
    k=i; 
    if(!fligou) k++;
    *k = a;
  }
  

  //  eventsList.sort(tmp1, eventsList.end(), SortEvent());

  //    cout<<"before adult clean:\n";
  //    Dump();


  // Adult events with age >= 1 or that are longer than StabTimescale
  for(tmp1=eventsList.begin(); tmp1 != eventsList.end(); tmp1++) {
    if(nowID - (*tmp1)->LastTimeID() >= 1) (*tmp1)->BeAdult();

    if((*tmp1)->LastTimeID() - (*tmp1)->FirstTimeID() >= maxClusterDuration ||
       (*tmp1)->GetSize() >= tidg->GetmaxClusterSize())
      (*tmp1)->BeMature();
  }

  //  cout<<"after adult clean:\n";
  //    Dump();

  /*
  system("rm -f OutFile.clust");
  for(tmp1=eventsList.begin(); tmp1 != eventsList.end(); tmp1++)
    (*tmp1)->AppendEventToFile("OutFile.clust");
  */
}



void EventsList::Purge(){ // remove Mature events and send them out

  if(eventsList.empty()) return;

  list<Event*>::iterator p;


  /* now done in Cluster
  // if event longer than stabScale, send it out by maturing it
  for(p = eventsList.begin(); p != eventsList.end();p++) {
    if((*p)->LastTimeID() - (*p)->FirstTimeID() >= maxClusterDuration) {
      (*p)->BeMature();
    }
  }
  */

  for(p = eventsList.begin(); p != eventsList.end();p++) {
    if((*p)->GetMaturity() == Mature) {

      /*
      if((*p)->GetChannelID() == 0)
	(*p)->AppendEventToFile("CleanData.clust");
      */

      tidg->SendEvent(*p);
      p = Erase(p); // points to next element

      if(p!=eventsList.begin())
	p--; // to avoid skipping
    }
  }
}


void EventsList::CPower () { // apply power threshold on mature events

  if(alpha >= 1.0) return; 

  float gammq(float, float);

  list<Event*>::iterator p;

  for(p = eventsList.begin(); p != eventsList.end();p++) {
    if((*p)->GetMaturity() == Mature) {
      
      float P = (*p)->ExcessPower(PThresh), // get total excess power
	DOF = (float)((*p)->GetSize());

      float prob = gammq(DOF,P); // get probability = integral from  P to infinity

      if(prob > alpha) { // we have a loser
	p = Erase(p);

	if(p==eventsList.begin()) break;
	p--;
      }
    }
  }      
}


void EventsList::CThresh () { // apply cluster thresholds

  list<Event*>::iterator p;

  /*
  for(p=eventsList.begin(); p!=eventsList.end(); p++)
    if((*p)->GetMaturity() == Adult) {
      if((*p)->GetSize() >= sigma) (*p)->BeMature();
      else p=Erase(p);
    }

  return;
  */

  p = eventsList.end(); // slightly more efficient to go backward
  p--;

  for(;;) {
    if((*p)->GetMaturity() == Adult) { // loop over adults

      // check raw size
      if((*p)->GetSize() >= sigma) {

	(*p)->BeMature(); // cluster large enough

      }
      else { // better have friends (delta search)
	bool FRIEND = 0;
	list<Event*>::iterator q; // only have to look backward

	if((*p)->GetSize() >= minSizeDelta &&
	   p != eventsList.begin()) {
	q = p; q--;

	while((*p)->FirstTimeID() - (*q)->LastTimeID() <= maxDt) {
	
	  if((*q)->GetMaturity() == Adult) {
	    if((*q)->GetSize() >= sigma) {

	      (*q)->BeMature();

	    }
	    else {
	      if(CloseEnough(*q, *p)) // merge q and p
		{**q += **p;
		
		(*q)->BeMature();
		Erase(p);
		if(p == eventsList.begin()) break;
		else p--;
		FRIEND = 1;
		break;}
	    }
	  }
	  if(q == eventsList.begin()) break;
	  else q--;
	}
	}

	if(!FRIEND && (nowID - (*p)->LastTimeID() > maxDt)) { // get rid of it
	  Erase(p);
	  if(p == eventsList.begin()) break;
	  else p--;
	}
      }
    }
    if(p == eventsList.begin()) break;
    else p--;
  }
}


bool EventsList::CloseEnough(Event *p, Event *q) { // delta test

  unsigned int s1, s2;
  
  if(p->GetSize() > q->GetSize()) {s1 = q->GetSize(); s2 = p->GetSize();}
  else {s2 = q->GetSize(); s1 = p->GetSize();}

  if(delta[s1-1][sigma-s2-1] > 1) 
    if((*p,*q) <= delta[s1-1][sigma-s2-1]) // test distance
      return 1;
  

  return 0;

}


void EventsList::SetNowId(Time tim) {

  if(firstTime == 0) firstTime = tim;


  //WARNING check
  unsigned int ntime = (unsigned int)ceil((tim - firstTime) / deltaT);


  if(nowID == 0) nowID = ntime;
  else {
    if(ntime > nowID + 1) {
      char buf[256];
      sprintf(buf,"%u",ntime-nowID-1);
      TidWarning(string("Dropping data segments for channel ") + chnname + string(". ") + buf + string(" seconds dropped."));
    }

    if(ntime <= nowID) {
      char buf[256];
      sprintf(buf,"%u",nowID-ntime+1);
      throw TidError(string("Repeated data segments for channel ") + chnname + string(". Data go back in time by ") + buf + string(" seconds."));
    }

    nowID = ntime;
  }


}
